from sagan.rtc import _parse_rtc_bytes


def to_bcd(x):
    return ((x // 10) << 4) | (x % 10)


def test_rtc_parse():
    for i in range(100):
        hundredths = i
        sec = i % 60
        min = i % 60
        hour = i % 24
        day = i % 31
        wk_day = i % 7
        mon = i % 12
        year = i

        regs = [
            to_bcd(hundredths),
            0x80 | to_bcd(sec),
            to_bcd(min),
            to_bcd(hour),
            to_bcd(day),
            to_bcd(wk_day),
            to_bcd(mon),
            to_bcd(year)
        ]

        parsed = _parse_rtc_bytes(regs)

        assert parsed.hundredths_of_seconds == hundredths
        assert parsed.second == sec
        assert parsed.minute == min
        assert parsed.hour == hour
        assert parsed.day == day
        assert parsed.week_day == wk_day
        assert parsed.month == mon
        assert parsed.year == year





class Leds:
    led1_pin = 27
    led2_pin = 22
    red_pin = 25
    green_pin = 23
    blue_pin = 24
    on = False
    off = True

    def __init__(self):
        pass

    def _set_pin(self, pin, value='on'):
        return
        # GPIO.output(pin, self.on if value == 'on' else self.off)

    def set_led1(self, *args):
        print("<[VIRTUAL MACHINE LOG] - led1 toggled>")
        return
        # self._set_pin(self.led1_pin, *args)

    def set_led2(self, *args):
        print("<[VIRTUAL MACHINE LOG] - led2 toggled>")
        return
        # self._set_pin(self.led2_pin, *args)

    def set_red(self, *args):
        print("<[VIRTUAL MACHINE LOG] - red led toggled>")
        return
        # self._set_pin(self.red_pin, *args)

    def set_green(self, *args):
        print("<[VIRTUAL MACHINE LOG] - green led toggled>")
        return
        # self._set_pin(self.green_pin, *args)

    def set_blue(self, *args):
        print("<[VIRTUAL MACHINE LOG] - blue led toggled>")
        return
        # self._set_pin(self.blue_pin, *args)




from sagan import temperature
from nose.tools import eq_


def test_temperature_parse():
    eq_(temperature._parse_temp_bytes(0b01111111, 0), 127.0)
    eq_(temperature._parse_temp_bytes(0b01111110, 0b11100000),  126.875)
    eq_(temperature._parse_temp_bytes(0b01111110, 0b00100000),  126.125)
    eq_(temperature._parse_temp_bytes(0b01111101, 0b00000000),  125.000)
    eq_(temperature._parse_temp_bytes(0b00011001, 0b00000000),   25.000)
    eq_(temperature._parse_temp_bytes(0b00000000, 0b00100000),    0.125)
    eq_(temperature._parse_temp_bytes(0b00000000, 0b00000000),    0.000)
    eq_(temperature._parse_temp_bytes(0b11111111, 0b11100000),   -0.125)
    eq_(temperature._parse_temp_bytes(0b11100111, 0b00000000),  -25.000)
    eq_(temperature._parse_temp_bytes(0b11001001, 0b00100000),  -54.875)
    eq_(temperature._parse_temp_bytes(0b11001001, 0b00000000),  -55.000)



from .i2c import I2cDevice
from collections import namedtuple


UvaMeasurement = namedtuple(
    'UvaMeasurement',
    'uva'
)


class UvaSensor(I2cDevice):
    def self_test(self):
        # There is nothing to do here, only readable registers are the output data
        return True

    def configure(self, args):
        pass
        #self.bus.write_byte(0x38, 0x60)

    def measure(self):
        """
        :return: UVa reading in W m^-2
        """
        # scale from sensor is 5 uW / cm^2  / encoder count.
        # msb = self.bus.read_byte_data(0x39, 0x00)
        # lsb = self.bus.read_byte_data(0x38, 0x00)
        # return UvaMeasurement(((msb << 8) | lsb) * 5e-2)
        return UvaMeasurement(0.0)

    @property
    def uva(self):
        return self.measure().uva



from .i2c import I2cDevice
from collections import namedtuple
from datetime import datetime
#
RtcTimeMeasurement = namedtuple('RtcTimeTuple', 'year month week_day day hour minute second hundredths_of_seconds')
#
#
# def _parse_rtc_bytes(time_regs):
#     hundredths_of_seconds = ((time_regs[0] & 0xF0) >> 4) * 10 + (time_regs[0] & 0x0F)
#     seconds = ((time_regs[1] & 0x70) >> 4) * 10 + (time_regs[1] & 0x0F)
#     minutes = ((time_regs[2] & 0x70) >> 4) * 10 + (time_regs[2] & 0x0F)
#     hours = ((time_regs[3] & 0x30) >> 4) * 10 + (time_regs[3] & 0x0F)
#     days = ((time_regs[4] & 0x30) >> 4) * 10 + (time_regs[4] & 0x0F)
#     week_day = time_regs[5] & 0x07
#     month = ((time_regs[6] & 0x10) >> 4) * 10 + (time_regs[6] & 0x0F)
#     year = ((time_regs[7] & 0xF0) >> 4) * 10 + (time_regs[7] & 0x0F)
#     return RtcTimeMeasurement(year, month, week_day, days, hours, minutes, seconds, hundredths_of_seconds)


class RealTimeClock(I2cDevice):
    def self_test(self):
        return True

    def configure(self, args):
        # self.write(0x28, [0x80])
        # self.write(0x25, [0x20])
        return

    def measure(self):
        """
        :return: Current real time clock values
        :rtype: RtcTimeTuple
        """
        #time_regs = self.read_and_unpack(0x00, 'B' * 8)
        date = datetime.now()
        return RtcTimeMeasurement(
            date.year, date.month, date.weekday(), date.day,date.hour,
            date.minute, date.second, date.microsecond)



from operator import add
from functools import reduce

from nose.tools import assert_almost_equal

from sagan.rgb_ir import _parse_rgb_ir_bytes


def encode_24bit_channel(value):
    return [value & 0xFFFF, value >> 16]


def encode_4_channel_bytes(values):
    return reduce(add, (encode_24bit_channel(channel) for channel in values), [])


def test_rgb_ir_parse():
    for i in range(10, 2 ** 24, 400000):
        inputs = (i * 0.1, i * 0.2, i * 0.3, i * 0.4)
        inputs = [int(x) for x in inputs]
        colour_bytes = encode_4_channel_bytes(inputs)
        total = sum(inputs)
        required = [x / total for x in inputs]
        actual = _parse_rgb_ir_bytes(colour_bytes)
        assert_almost_equal(actual[0], required[3])
        assert_almost_equal(actual[1], required[1])
        assert_almost_equal(actual[2], required[2])
        assert_almost_equal(actual[3], required[0])



from .i2c import I2cDevice
from collections import namedtuple

# # LSM9DS0 Gyro Registers
# CTRL_REG1_G = 0x20
# CTRL_REG2_G = 0x21
# CTRL_REG3_G = 0x22
# CTRL_REG4_G = 0x23
# CTRL_REG5_G = 0x24
#
# # LSM9DS0 Accel and Magneto Registers
# CTRL_REG1_XM = 0x20
# CTRL_REG2_XM = 0x21
# CTRL_REG3_XM = 0x22
# CTRL_REG4_XM = 0x23
# CTRL_REG5_XM = 0x24
# CTRL_REG6_XM = 0x25
# CTRL_REG7_XM = 0x26


AccelerometerMeasurement = namedtuple(
    'AccelerometerMeasurement',
    'x y z'
)


GryoscopeMeasurement = namedtuple(
    'GryoscopeMeasurement',
    'x y z'
)


MagnetometerMeasurement = namedtuple(
    'MagnetometerMeasurement',
    'x y z'
)


class Lsm9ds0I2cDevice(I2cDevice):
    """
    This overrides the read method to toggle the high bit in the register address.
    This is needed for multi-byte reads.
    """
    def read(self, cmd, length):
        return 0.0


class Accelerometer(Lsm9ds0I2cDevice):
    # These values come from the LSM9DS0 data sheet p13 table3 in the row about sensitivities.
    acceleration_scale = 0.000732 * 9.80665
    magnetometer_scale = 0.00048

    def configure(self, args):
        pass

    def measure(self):
        """
        :return: acceleration (X, Y, Z triple in m s^-1)
        """
        return AccelerometerMeasurement(0.0, 0.0, 0.0)

    @property
    def x(self):
        return 0.0

    @property
    def y(self):
        return 0.0

    @property
    def z(self):
        return 9.8


class Magnetometer(Lsm9ds0I2cDevice):
    # These values come from the LSM9DS0 data sheet p13 table3 in the row about sensitivities.
    acceleration_scale = 0.000732 * 9.80665
    magnetometer_scale = 0.00048

    def self_test(self):
        return True

    def configure(self, args):
        return

    def measure(self):
        """
        :return: magnetic field (X, Y, Z triple in mgauss)
        """
        return MagnetometerMeasurement(0.0, 0.0, 0.0)

    @property
    def x(self):
        return 0.0

    @property
    def y(self):
        return 0.0

    @property
    def z(self):
        return 0.0


class Gyroscope(Lsm9ds0I2cDevice):
    gyroscope_scale = 0.070

    def self_test(self):
        return True

    def configure(self, args):
        return

    def measure(self):
        """
        :return: X, Y, Z triple in degrees per second
        """
        return GryoscopeMeasurement(0.0, 0.0, 0.0)

    @property
    def x(self):
        return 0.0

    @property
    def y(self):
        return 0.0

    @property
    def z(self):
        return 0.0


from .i2c import I2cDevice
from collections import namedtuple


TemperatureMeasurement = namedtuple(
    'TemperatureMeasurement',
    'temperature'
)

def _parse_temp_bytes(hi, lo):
    # value = ((hi << 8) | lo) >> 5
    # if hi & (1 << 7):
    #     value = - ((~value & ((1 << 11) - 1)) + 1)
    # value *= 0.125
    return 0.0


class TemperatureSensor(I2cDevice):
    """
    Interface for LM75B.
    """
    def measure(self):
        """
        :return: Temperature reading in C
        """
        # temp = self.read(0, 2)
        return TemperatureMeasurement(0.0)

    def self_test(self):
        # not sure what self test to do.
        # read and write a register?
        return True

    @property
    def temperature(self):
        return self.measure().temperature




from .i2c import I2cDevice
from collections import namedtuple


RgbIrMeasurement = namedtuple(
    'RgbIrMeasurement',
    'red green blue ir'
)


def _parse_rgb_ir_bytes(colour_data):
    return 0.0, 0.0, 0.0, 0.0


class RgbIrSensor(I2cDevice):
    def self_test(self):
        id = self.read(0x06, 1)[0]
        return id == 0xB2

    def configure(self, args):
        # set light sensor enabled, colour sensing mode.
        self.pack_and_write(0x00, 'B', 0b00000110)
        super().configure(args)

    def measure(self):
        """
        :return: R, G, B and IR Channel readings and a fraction of the total.
        """
        return RgbIrMeasurement(*_parse_rgb_ir_bytes(0.0))

    @property
    def red(self):
        return self.measure()[0]

    @property
    def green(self):
        return self.measure()[1]

    @property
    def blue(self):
        return self.measure()[2]

    @property
    def ir(self):
        return self.measure()[3]



import time
from .i2c import I2cDevice
from collections import namedtuple

BarometerMeasurement = namedtuple(
    'BarometerMeasurement',
    'temperature pressure humidity'
)

class Barometer(I2cDevice):

    """
    Interface for BME280 pressure and humidity
    """
    def read_raw_measurements(self):
        return 0.0, 0.0, 0.0

    def apply_calibration(self, t_raw, p_raw, h_raw, t_calib=None):
        pass
        return 0.0,0.0,0.0

    def measure(self):
        """
        :return: tuple of: temperature (C), pressure (Pa), humidity (% relative humidity)
        """
        return 0.0, 0.0, 0.0


    @property
    def temperature(self):
        return 0.0

    @property
    def pressure(self):
        return 0.0

    @property
    def humidity(self):
        return 0.0



from distutils.core import setup

setup(
    name='saganvm',
    version='1.1.5',
    packages=['sagan'],
    url='',
    license='',
    author='T A H Smith, A W Collins',
    author_email='saganvirtual@cuberider.com',
    description='Python library for interfacing with launchpad website')



"""
This test uses cases generated with the following C language code:

    #include <stdio.h>
    #include <stdint.h>

    typedef int32_t BME280_S32_t;
    typedef uint32_t BME280_U32_t;

    BME280_U32_t dig_T1 = 28353;
    BME280_S32_t dig_T2 = 26575;
    BME280_S32_t dig_T3 = 50;

    BME280_U32_t dig_P1 = 37446;
    BME280_S32_t dig_P2 = -10740;
    BME280_S32_t dig_P3 = 3024;
    BME280_S32_t dig_P4 = 8986;
    BME280_S32_t dig_P5 = -124;
    BME280_S32_t dig_P6 = -7;
    BME280_S32_t dig_P7 = 9900;
    BME280_S32_t dig_P8 = -10230;
    BME280_S32_t dig_P9 = 4285;

    BME280_U32_t dig_H1 = 0;
    BME280_S32_t dig_H2 = 354;
    BME280_U32_t dig_H3 = 0;
    BME280_S32_t dig_H4 = 340;
    BME280_S32_t dig_H5 = 0;
    BME280_S32_t dig_H6 = 30;

    // Returns temperature in DegC, double precision. Output value of “51.23” equals
    // 51.23 DegC. // t_fine carries fine temperature as global value
    BME280_S32_t t_fine;
    double BME280_compensate_T_double(BME280_S32_t adc_T) {
      double var1, var2, T;
      var1 = (((double)adc_T) / 16384.0 - ((double)dig_T1) / 1024.0) *
             ((double)dig_T2);
      var2 = ((((double)adc_T) / 131072.0 - ((double)dig_T1) / 8192.0) *
              (((double)adc_T) / 131072.0 - ((double)dig_T1) / 8192.0)) *
             ((double)dig_T3);
      t_fine = (BME280_S32_t)(var1 + var2);
      T = (var1 + var2) / 5120.0;
      return T;
    }
    // Returns pressure in Pa as double. Output value of “96386.2” equals 96386.2 Pa
    // = 963.862 hPa
    double BME280_compensate_P_double(BME280_S32_t adc_P) {
      double var1, var2, p;
      var1 = ((double)t_fine / 2.0) - 64000.0;
      var2 = var1 * var1 * ((double)dig_P6) / 32768.0;
      var2 = var2 + var1 * ((double)dig_P5) * 2.0;
      var2 = (var2 / 4.0) + (((double)dig_P4) * 65536.0);
      var1 = (((double)dig_P3) * var1 * var1 / 524288.0 + ((double)dig_P2) * var1) /
             524288.0;
      var1 = (1.0 + var1 / 32768.0) * ((double)dig_P1);
      if (var1 == 0.0) {
        return 0;  // avoid exception caused by division by zero
      }
      p = 1048576.0 - (double)adc_P;
      p = (p - (var2 / 4096.0)) * 6250.0 / var1;
      var1 = ((double)dig_P9) * p * p / 2147483648.0;
      var2 = p * ((double)dig_P8) / 32768.0;
      p = p + (var1 + var2 + ((double)dig_P7)) / 16.0;
      return p;
    }
    // Returns humidity in %rH as as double. Output value of “46.332” represents
    // 46.332 %rH
    double bme280_compensate_H_double(BME280_S32_t adc_H)
    {
      double var_H;
      var_H = (((double)t_fine) - 76800.0);
      var_H =
          (adc_H - (((double)dig_H4) * 64.0 + ((double)dig_H5) / 16384.0 * var_H)) *
          (((double)dig_H2) / 65536.0 *
           (1.0 +
            ((double)dig_H6) / 67108864.0 * var_H *
                (1.0 + ((double)dig_H3) / 67108864.0 * var_H)));
      var_H = var_H * (1.0 - ((double)dig_H1) * var_H / 524288.0);
      if (var_H > 100.0)
        var_H = 100.0;
      else if (var_H < 0.0)
        var_H = 0.0;
      return var_H;
    }

    void evaluate_curve(const char* name, double (*curve_fn) (BME280_S32_t), BME280_S32_t a, BME280_S32_t b)
    {
        printf("%s_values = (", name);
        for(int i = 0; i < 10; i += 1)
        {
            double t = (*curve_fn)(i * a + b);
            printf("%g, ", t);
        }
        printf(")\n");
    }

    int main(int argc, char** argv)
    {
        evaluate_curve("t", &BME280_compensate_T_double, 100000, 0);
        evaluate_curve("p", &BME280_compensate_P_double, 100000, 0);
        evaluate_curve("h", &bme280_compensate_H_double, 100000, 0);
    }

"""

from sagan.baro import Barometer


def assert_almost_equal(x, y, rel_tolerance=1e-4):
    if x == y == 0:
        return
    rel_diff = 2 * abs(x - y) / (x + y)
    assert rel_diff < rel_tolerance, "{0} != {1}, with relative tolerance {2}".format(x, y, rel_tolerance)


def test_calibration_curves():
    baro = Barometer(None, 0x00)
    baro.temperature_parameters[0] = 28353
    baro.temperature_parameters[1] = 26575
    baro.temperature_parameters[2] = 50

    baro.pressure_parameters[0] = 37446
    baro.pressure_parameters[1] = -10740
    baro.pressure_parameters[2] = 3024
    baro.pressure_parameters[3] = 8986
    baro.pressure_parameters[4] = -124
    baro.pressure_parameters[5] = -7
    baro.pressure_parameters[6] = 9900
    baro.pressure_parameters[7] = -10230
    baro.pressure_parameters[8] = 4285

    baro.humidity_parameters[0] = 0
    baro.humidity_parameters[1] = 354
    baro.humidity_parameters[2] = 0
    baro.humidity_parameters[3] = 340
    baro.humidity_parameters[4] = 0
    baro.humidity_parameters[5] = 30

    # The following lines are output by the C program at the start of this module
    t_a = 100000
    t_b = -100000
    t_values = (-175.221, -143.598, -111.964, -80.3188, -48.6621, -16.994, 14.6855, 46.3763, 78.0785, 109.792,)
    p_a = 100000
    p_b = -100000
    p_values = (168562, 151534, 134575, 117685, 100865, 84114.7, 67433.7, 50822.1, 34280, 17807.5,)
    h_a = 2100
    h_b = 21000
    h_values = (0, 7.40383, 19.0068, 30.6099, 42.2129, 53.8159, 65.4189, 77.0219, 88.6249, 100,)

    for i in range(10):
        t_raw = i * t_a + t_b
        p_raw = i * p_a + p_b
        h_raw = i * h_a + h_b
        t, p, h = baro.apply_calibration(t_raw, p_raw, h_raw, 25.0 * 5120)

        assert_almost_equal(t, t_values[i])
        assert_almost_equal(p, p_values[i])
        assert_almost_equal(h, h_values[i])



from collections import namedtuple
from datetime import datetime

CAPTURE_EXECUTABLE = 'ov2640_capture'
X_RESOLUTION = 800
Y_RESOLUTION = 600

CameraCaptureResult = namedtuple(
    'CameraCaptureResult',
    'filename'
)

class Camera:
    def capture(self, filename=None):
        print("<[VIRTUAL MACHINE LOG] - Picture taken>")




import time
from .baro import Barometer
from .temperature import TemperatureSensor
from .imu import Accelerometer, Magnetometer, Gyroscope
from .rgb_ir import RgbIrSensor
from .uva import UvaSensor
from .rtc import RealTimeClock
from .leds import Leds
from .arducam import Camera

barometer = Barometer(None, 0x76)
bottom_thermometer = TemperatureSensor(None, 0x48)
top_thermometer = TemperatureSensor(None, 0x49)
accelerometer = Accelerometer(None, 0x1D)
magnetometer = Magnetometer(None, 0x1D)
gyroscope = Gyroscope(None, 0x6b)
rgb_ir_sensor = RgbIrSensor(None, 0x52)
uva_sensor = UvaSensor(None, 0x38)
real_time_clock = RealTimeClock(None, 0x51)

sensors = [
    barometer,
    bottom_thermometer,
    top_thermometer,
    accelerometer,
    gyroscope,
    rgb_ir_sensor,
    uva_sensor,
    real_time_clock
]

for sensor in sensors:
    sensor.configure({})

leds = Leds()
camera = Camera()


__all__ = (
    'barometer',
    'bottom_thermometer',
    'top_thermometer',
    'accelerometer',
    'gyroscope',
    'magnetometer',
    'rgb_ir_sensor',
    'uva_sensor',
    'real_time_clock',
    'leds',
    'camera'
)


def test():
    print("RUNNING TEST...")
    try:

        print(barometer.measure())
        print(bottom_thermometer.measure())
        print(top_thermometer.measure())
        print(accelerometer.measure())
        print(rgb_ir_sensor.measure())
        print(uva_sensor.measure())
        print(real_time_clock.measure())
        leds.set_blue("off")
        leds.set_green("off")
        leds.set_red("off")
        leds.set_led1("off")
        leds.set_led2("off")
        time.sleep(0.25)
        leds.set_led1("on")
        time.sleep(0.25)
        leds.set_led1("off")
        leds.set_led2("on")
        time.sleep(0.25)
        leds.set_led2("off")
        leds.set_red("on")
        time.sleep(0.25)
        leds.set_red("off")
        leds.set_green("on")
        time.sleep(0.25)
        leds.set_green("off")
        leds.set_blue("on")
        time.sleep(0.25)
        leds.set_blue("off")
    except:
        print("ERROR")
        return
    print("PASS")









from abc import abstractmethod
import struct


class I2cDevice:
    """
    A light wrapper on top of smbus for convenience.
    """
    def __init__(self, bus, address):
        self.bus = bus
        self.address = address

    def read(self, cmd, length):
        return

    def write(self, cmd, values):
        return

    def read_and_unpack(self, cmd, fmt):
        return

    def pack_and_write(self, cmd, fmt, values):
        return

    @abstractmethod
    def configure(self, args):
        pass

    @abstractmethod
    def self_test(self):
        pass

    def __repr__(self):
        return ""
